"""
apkg packaging file cache
"""

import hashlib
import json
from pathlib import Path

from apkg.log import getLogger


log = getLogger(__name__)


def file_checksum(path):
    chsum = hashlib.sha256()
    with Path(path).open("rb") as f:
        for chunk in iter(lambda: f.read(4096), b""):
            chsum.update(chunk)
    return chsum.hexdigest()[:20]


class ProjectCache:
    def __init__(self, project):
        self.project = project
        self.loaded = False
        self.cache = {}
        self.checksum = None

    def save(self):
        json.dump(self.cache, self.project.cache_path.open('w'))

    def load(self):
        cache_path = self.project.cache_path
        if not cache_path.exists():
            log.verbose("cache not found: %s", cache_path)
            return
        log.verbose("loading cache: %s", cache_path)
        self.cache = json.load(self.project.cache_path.open('r'))

    def _ensure_load(self):
        """
        ensure cache is loaded on demand and only once

        you don't need to call this directly
        """
        if self.loaded:
            return
        self.load()
        self.loaded = True

    def update(self, cache_name, key, paths):
        """
        update cache entry
        """
        log.verbose("cache update for %s: %s -> %s",
                    cache_name, key, paths[0])
        assert key
        self._ensure_load()
        if cache_name not in self.cache:
            self.cache[cache_name] = {}
        entries = list(map(path2entry, paths))
        self.cache[cache_name][key] = entries
        self.save()

    def get(self, cache_name, key):
        """
        get cache entry or None
        """
        log.verbose("cache query for %s: %s",
                    cache_name, key)

        def validate(path, checksum):
            if not path.exists():
                log.info("removing missing file from cache: %s", path)
                self.delete(cache_name, key)
                return False
            real_checksum = file_checksum(path)
            if real_checksum != checksum:
                log.info("removing invalid cache entry: %s", path)
                self.delete(cache_name, key)
                return False
            return True

        def entry2path_valid(e):
            return entry2path(e, validate_fun=validate)

        assert key
        self._ensure_load()
        entries = self.cache.get(cache_name, {}).get(key)
        if not entries:
            return None
        paths = list(map(entry2path_valid, entries))
        if None in paths:
            # invalid entry
            return None
        return paths

    def delete(self, cache_name, key):
        """
        delete cache entry
        """
        self.cache[cache_name].pop(key, None)
        self.save()

    def enabled(self, use_cache=True):
        """
        helper to tell and log if caching is enabled and supported

        optional use_cache argument provided for shared
        argument parsing and logging from apkg.commands
        """
        if use_cache:
            vcs = self.project.vcs
            if vcs:
                log.verbose("%s VCS detected -> cache ENABLED", vcs)
                return True
            else:
                log.verbose("VCS not detected -> cache DISABLED")
        else:
            log.verbose("cache DISABLED")
        return False


def path2entry(path):
    """
    convert a path to corresponding cache entry

    return (fn, checksum) or a list of that on multiple paths
    """
    return str(path), file_checksum(path)


def entry2path(entry, validate_fun=None):
    """
    convert cache entry to a corresponding path

    if validate_fun is specified, it's used confirm file has
    valid checksum and flush invalid cache entry if it doesn't
    """
    fn, checksum = entry
    p = Path(fn)
    if validate_fun:
        if not validate_fun(p, checksum):
            return None
    return p



import click

from apkg import adistro
from apkg import ex
from apkg.log import getLogger
from apkg import pkgstyle


log = getLogger(__name__)


@click.command(name='system-setup')
@click.option('-I', '--isolated', is_flag=True,
              help="also install packages for isolated build")
@click.option('-d', '--distro',
              help="override target distro  [default: current]")
@click.option('--ask/--no-ask', 'interactive',
              default=False, show_default=True,
              help="enable/disable interactive mode")
def cli_system_setup(*args, **kwargs):
    """
    setup system for packaging
    """
    return system_setup(*args, **kwargs)


def system_setup(
        isolated=False,
        distro=None,
        interactive=False):
    """
    setup system for packaging

    Install native distro packages required for packaging.
    """
    log.bold("system setup for packaging")

    distro = adistro.distro_arg(distro)
    log.info("target distro: %s", distro)

    style = pkgstyle.get_pkgstyle_for_distro(distro)
    if not style:
        raise ex.DistroNotSupported(distro=distro)
    log.info("target pkgstyle: %s", style.name)

    distro_reqs = getattr(style, 'DISTRO_REQUIRES', {})
    reqs = distro_reqs.get('core', [])
    if isolated:
        reqs += distro_reqs.get('isolated', [])
    if reqs:
        pkgstyle.call_pkgstyle_fun(
            style, 'install_distro_packages',
            reqs,
            interactive=interactive)
    else:
        log.info("no distro packages required")

    log.success("system ready for packaging")


APKG_CLI_COMMANDS = [cli_system_setup]



from pathlib import Path

import click

from apkg import ex
from apkg.cli import cli
from apkg.util import common
from apkg.log import getLogger
from apkg.project import Project
from apkg.util.run import run
import apkg.util.shutil35 as shutil


log = getLogger(__name__)


@cli.command(name='make-archive', aliases=['ar'])
@click.option('-O', '--result-dir',
              help="put results into specified dir")
@click.option('--cache/--no-cache', default=True, show_default=True,
              help="enable/distable cache")
@click.help_option('-h', '--help', help='show this help')
def cli_make_archive(*args, **kwargs):
    """
    create dev archive from current project state
    """
    results = make_archive(*args, **kwargs)
    common.print_results(results)
    return results


def make_archive(
        result_dir=None,
        cache=True,
        project=None):
    """
    create dev archive from current project state

    Use script specified by project.make_archive_script config option.
    """
    log.bold("creating dev archive")
    proj = project or Project()

    use_cache = proj.cache.enabled(cache)
    if use_cache:
        cache_name = 'archive/dev'
        cache_key = proj.checksum
        cached = common.get_cached_paths(
            proj, cache_name, cache_key, result_dir)
        if cached:
            log.success("reuse cached archive: %s", cached[0])
            return cached

    script = proj.config_get('project.make_archive_script')
    if not script:
        msg = ("make-archive requires project.make_archive_script option to\n"
               "be set in project config to a script that creates project\n"
               "archive and prints its path to stdout.\n\n"
               "Please update project config with required information:\n\n"
               "%s" % proj.config_path)
        raise ex.MissingRequiredConfigOption(msg=msg)

    log.info("running make_archive_script: %s", script)
    out = run(script)
    # last script stdout line is expected to be path to resulting archive
    _, _, last_line = out.rpartition('\n')
    in_archive_path = Path(last_line)
    if not in_archive_path.exists():
        msg = ("make_archive_script finished successfully but the archive\n"
               "(indicated by last script stdout line) doesn't exist:\n\n"
               "%s" % in_archive_path)
        raise ex.UnexpectedCommandOutput(msg=msg)
    log.info("archive created: %s", in_archive_path)

    if result_dir:
        ar_base_path = Path(result_dir)
    else:
        ar_base_path = proj.dev_archive_path
    archive_fn = in_archive_path.name
    archive_path = ar_base_path / archive_fn
    if archive_path != in_archive_path:
        log.info("copying archive to: %s", archive_path)
        ar_base_path.mkdir(parents=True, exist_ok=True)
        shutil.copy(in_archive_path, archive_path)
    log.success("made archive: %s", archive_path)
    results = [archive_path]
    if use_cache:
        proj.cache.update(
            cache_name, cache_key, results)
    return results


APKG_CLI_COMMANDS = [cli_make_archive]



"""
distro-specific apkg helpers related to current distro

current distro info isn't going to change during run so cache that
"""
import re

import distro


def distro_arg(distro_str):
    """
    normalize --distro argument as used by CLI

    return current distro if not set
    """
    if distro_str:
        # convert custom distro string to idver format
        return distro2idver(distro_str)
    # use current distro by default
    return idver()


def distro2idver(distro_name):
    """
    convert generic distro string into idver format
    """
    return re.sub(r'\s+', '-', distro_name.strip().lower())


def idver():
    """
    return current distro in default idver format
    used by apkg to reference individual distros

    examples: debian-10, fedora-32, arch
    """
    global _idver
    if _idver is None:
        _idver = id()
        ver = version()
        if ver:
            _idver += "-%s" % (ver)
    return _idver


def fullname():
    """
    return human readable string describing current distro
    """
    global _fullname
    parts = [
        name(),
        distro.version(pretty=True),
    ]
    return " ".join([p for p in parts if p])


def id():  # pylint: disable=redefined-builtin
    global _id
    if _id is None:
        _id = distro.id()
    return _id


def name():
    global _name
    if _name is None:
        _name = distro.name()
    return _name


def codename():
    global _codename
    if _codename is None:
        _codename = distro.codename()
    return _codename


def version():
    global _version
    if _version is None:
        _version = distro.version()
    return _version


_id = None
_idver = None
_name = None
_codename = None
_version = None
_fullname = None



import glob
import hashlib
from pathlib import Path
import os
import re

import jinja2
import toml
try:
    from functools import cached_property
except ImportError:
    from cached_property import cached_property

from apkg import cache as _cache
from apkg import ex
from apkg.log import getLogger
from apkg import pkgtemplate
from apkg.util.git import git
from apkg.util import upstreamversion


log = getLogger(__name__)


INPUT_BASE_DIR = 'distro'
OUTPUT_BASE_DIR = 'pkg'
CONFIG_FN = 'apkg.toml'


# pylint: disable=too-many-instance-attributes
# TODO: consider moving paths to project.path.* to address this warning
class Project:
    """
    Project class serves as high level interface to projecs in need of
    packaging
    """
    config = {}

    name = None
    path = None
    templates_path = None
    cache_path = None
    config_base_path = None
    config_path = None
    archive_path = None
    dev_archive_path = None
    upstream_archive_path = None
    unpacked_archive_path = None
    build_path = None
    package_build_path = None
    srcpkg_build_path = None
    package_out_path = None
    srcpkg_out_path = None

    def __init__(self, path=None, autoload=True):
        if path:
            self.path = path
        else:
            self.path = Path('.')
        if autoload:
            self.load()
        self.cache = _cache.ProjectCache(self)

    def update_attrs(self):
        """
        update project attributes based on current config
        """
        self.name = self.config_get('project.name')
        if self.name:
            log.verbose("project name from config: %s", self.name)
        else:
            self.name = self.path.resolve().name
            log.verbose("project name not in config - "
                        "guessing from path: %s", self.name)

    def update_paths(self):
        """
        fill in projects paths based on current self.path and self.config
        """
        # package templates: distro/pkg
        self.templates_path = self.input_path / 'pkg'
        # archives: pkg/archives/{dev,upstream,unpacked}
        self.archive_path = self.output_path / 'archives'
        self.dev_archive_path = self.archive_path / 'dev'
        self.upstream_archive_path = self.archive_path / 'upstream'
        self.unpacked_archive_path = self.archive_path / 'unpacked'
        # build: pkg/build/{src-,}pkg
        self.build_path = self.output_path / 'build'
        self.package_build_path = self.build_path / 'pkgs'
        self.srcpkg_build_path = self.build_path / 'srcpkgs'
        # output: pkg/{src-,}pkg
        self.package_out_path = self.output_path / 'pkgs'
        self.srcpkg_out_path = self.output_path / 'srcpkgs'
        # cache: pkg/.cache.json
        self.cache_path = self.output_path / '.cache.json'

    def load(self,
             input_path=None,
             output_path=None):
        """
        load project config and update its attributes
        """
        if not input_path:
            input_path = self.path / INPUT_BASE_DIR
        self.input_path = input_path

        if not output_path:
            output_path = self.path / OUTPUT_BASE_DIR
        self.output_path = output_path

        self.config_base_path = self.input_path / 'config'
        self.config_path = self.config_base_path / CONFIG_FN
        self.load_config()
        self.update_attrs()
        self.update_paths()

    def load_config(self):
        if self.config_path.exists():
            log.verbose("loading project config: %s", self.config_path)
            self.config = toml.load(self.config_path.open())
            return True
        else:
            log.verbose("project config not found: %s", self.config_path)
            return False

    def config_get(self, option):
        """
        get config option if set or None

        example options: 'project.name', 'upstream.archive_url'
        """
        c = self.config
        for key in option.split('.'):
            try:
                c = c[key]
            except KeyError:
                return None
        return c

    @cached_property
    def vcs(self):
        """
        Version Control System used in project

        possible outputs: 'git', None
        """
        o = git('rev-parse', silent=True, fatal=False)
        if o.return_code == 0:
            return 'git'
        return None

    @cached_property
    def checksum(self):
        """
        checksum of current project state

        requires VCS (git), only computed once
        """
        if self.vcs == 'git':
            checksum = git.current_commit()[:10]
            diff = git('diff', log_cmd=False)
            if diff:
                diff_hash = hashlib.sha256(diff.encode('utf-8'))
                checksum += '-%s' % diff_hash.hexdigest()[:10]
            return checksum
        return None

    def upstream_archive_url(self, version):
        url = self.config_get('upstream.archive_url')
        if not url:
            return None
        env = {'project': self, 'version': version}
        url = jinja2.Template(url).render(**env)
        return url

    def upstream_signature_url(self, version):
        url = self.config_get('upstream.signature_url')
        if not url:
            return None
        env = {'project': self, 'version': version}
        url = jinja2.Template(url).render(**env)
        return url

    @cached_property
    def upstream_version(self):
        """
        check latest upstream version

        upstream is only queried once

        possible outputs: version, None
        """
        uv_script = self.config_get('upstream.version_script')
        if uv_script:
            v = upstreamversion.version_from_script(
                uv_script, script_name='upstream.version_script')
            log.info("detected upstream version (from script): %s", v)
            return v
        ar_url = self.upstream_archive_url('VERSION')
        if ar_url:
            m = re.match(r'(.*/)[^/]+', ar_url)
            ar_base_url = m.group(1)
            v = upstreamversion.version_from_listing(ar_base_url)
            log.info("detected upstream version: %s", v)
            return v
        return None

    @cached_property
    def templates(self):
        if self.templates_path.exists():
            return load_templates(self.templates_path)
        else:
            return []

    def _get_template_for_distro(self, distro):
        for t in self.templates:
            ps = t.pkgstyle
            for d in ps.SUPPORTED_DISTROS:
                # NOTE: this is very simplistic
                if d in distro:
                    return t
        return None

    def get_template_for_distro(self, distro):
        ldistro = distro.lower()
        template = self._get_template_for_distro(ldistro)
        if not template:
            tdir = self.templates_path
            msg = ("missing package template for distro: %s\n\n"
                   "you can add it into: %s" % (distro, tdir))
            raise ex.MissingPackagingTemplate(msg=msg)
        return template

    def find_archives_by_name(self, name, upstream=False):
        """
        find archive files with supplied name in expected project paths
        """
        if upstream:
            ar_path = self.upstream_archive_path
        else:
            ar_path = self.dev_archive_path
        return glob.glob("%s/%s*" % (ar_path, name))


def load_templates(path):
    templates = []
    for entry_path in glob.glob('%s/*' % path):
        if os.path.isdir(entry_path):
            template = pkgtemplate.PackageTemplate(entry_path)
            if template.pkgstyle:
                templates.append(template)
            else:
                log.warning("ignoring unknown package style in %s", entry_path)
    return templates



"""
apkg package style for **Debian**
and its many clones such as Ubuntu or Mint.

**source template**: content of `debian/` dir (`control`, `changelog`, ...)

**source package:** `*.dsc` + archives

**packages:** `*.deb`

**required distro packages**:

 * core: `devscripts`
 * isolated build: `pbuilder`
"""
import glob
import os
from pathlib import Path
import re
import sys
import tempfile

from apkg import ex
from apkg.log import getLogger
from apkg import parse
from apkg import pkgtemplate
from apkg.util.run import cd, run, sudo
from apkg.util.archive import unpack_archive
import apkg.util.shutil35 as shutil


log = getLogger(__name__)


SUPPORTED_DISTROS = [
    "ubuntu",
    "debian",
    "linuxmint",
    "raspbian",
]
DISTRO_REQUIRES = {
    'core': ['devscripts'],
    'isolated': ['pbuilder'],
}


RE_PKG_NAME = r'Source:\s*(\S+)'
# orbital regexp cannon to parse Build-Depends from debian/control
RE_BUILD_DEPENDS = (
    r'(?:\n|\A)Build-Depends(?:-Indep)?:[ \t]*'  # no whitespace before
    r'(?:\n[ \t]+)?'  # optional leading newline with whitespace
    r'((?:[^,\n]+)'   # first build dep
    r'(?:,(?:[ \t]*'  # comma separator and optional whitespace
    r'(?:\n[ \t]+)?'  # optional newline starting with whitespace
    r'[^,\n]+))*)'    # 0-N other build deps
)


def is_valid_template(path):
    deb_files = ['rules', 'control', 'changelog']
    return all((path / f).exists() for f in deb_files)


def get_template_name(path):
    control = path / 'control'

    for line in control.open():
        m = re.match(RE_PKG_NAME, line)
        if m:
            return m.group(1)

    raise ex.ParsingFailed(
        msg="unable to determine Source from: %s" % control)


def get_srcpkg_nvr(path):
    nvr, _, _ = str(path.name).rpartition('.')
    return nvr


def copy_srcpkg_files(src_path, dst_path):
    # not part of pkgstyle interface yet but probably should be
    for pattern in [
            '*.dsc',
            '*_source.*',  # questionable, some tools need these
            '*.debian.tar.*',
            '*.orig.tar.*',
            '*.diff.*']:
        for f in glob.iglob('%s/%s' % (src_path, pattern)):
            srcp = Path(f)
            shutil.copyfile(f, dst_path / srcp.name)


def build_srcpkg(
        build_path,
        out_path,
        archive_paths,
        template,
        env):
    """
    build debian source package
    """
    archive_path = archive_paths[0]
    nv, _ = parse.split_archive_ext(archive_path.name)
    log.info("building deb source package: %s", nv)
    log.info("unpacking archive: %s", archive_path)
    source_path = unpack_archive(archive_path, build_path)
    log.verbose("source package root dir: %s", source_path)
    if not source_path or not source_path.exists():
        msg = "archive unpack didn't result in expected dir: %s" % source_path
        raise ex.UnexpectedCommandOutput(msg=msg)
    # render template
    debian_path = source_path / 'debian'
    template.render(debian_path, env)
    # copy archive with debian .orig name
    _, _, _, ext = parse.split_archive_fn(archive_path.name)
    debian_ar = "%s_%s.orig%s" % (env['name'], env['version'], ext)
    debian_ar_path = build_path / debian_ar
    log.info("copying archive into source package: %s", debian_ar_path)
    shutil.copyfile(archive_path, debian_ar_path)

    log.info("building deb source-only package...")
    with cd(source_path):
        run('dpkg-buildpackage',
            '-S',   # source-only, no binary files
            '-sa',  # source includes orig, always
            '-d',   # do not check build dependencies and conflicts
            '-nc',  # do not pre clean source tree
            '-us',  # unsigned source package.
            '-uc',  # unsigned .changes file.
            direct='auto')

    log.info("copying source package to result dir: %s", out_path)
    out_path.mkdir(parents=True)
    copy_srcpkg_files(build_path, out_path)
    fns = glob.glob('%s/*' % out_path)
    # make sure .dsc is first
    for i, fn in enumerate(fns):
        if fn.endswith('.dsc'):
            fns = [fns.pop(i)] + fns
            break
    else:
        raise ex.UnexpectedCommandOutput(
            msg="no *.dsc found after moving built source package")
    return list(map(Path, fns))


def build_packages(
        build_path,
        out_path,
        srcpkg_paths,
        **kwargs):
    """
    build .deb packages from source package
    """
    srcpkg_path = srcpkg_paths[0]
    build_path.mkdir(parents=True)
    out_path.mkdir(parents=True)
    isolated = kwargs.get('isolated')
    if isolated:
        log.info("starting isolated build using pbuilder")
        # TODO: ensure pbuilder's base image exists (pbuilder create)
        sudo('pbuilder', 'build',
             '--buildresult', build_path,
             srcpkg_path,
             preserve_env=True,  # preserve env inc. DEB_BUILD_OPTIONS
             direct='auto')
    else:
        # unpack source package
        log.info("unpacking source package for direct host build")
        srcpkg_abspath = srcpkg_path.resolve()
        with cd(build_path):
            run('dpkg-source', '-x', srcpkg_abspath,
                direct='auto')
        # find unpacked source dir
        try:
            source_glob = '%s/*/' % build_path
            source_path = Path(glob.glob(source_glob)[0])
        except IndexError:
            msg = "failed to find unpacked source dir: %s"
            raise ex.UnexpectedCommandOutput(msg % source_glob)

        log.info("starting direct host build using dpkg-buildpackage")
        with cd(source_path):
            # build
            run('dpkg-buildpackage',
                '-us',  # unsigned source package.
                '-uc',  # unsigned .changes file.
                direct='auto')

    pkgs = []
    log.info("copying built packages to result dir: %s", out_path)
    for src_pkg in glob.iglob('%s/*.deb' % build_path):
        dst_pkg = out_path / Path(src_pkg).name
        shutil.copyfile(src_pkg, dst_pkg)
        pkgs.append(dst_pkg)

    return pkgs


def install_distro_packages(
        packages,
        **kwargs):
    interactive = kwargs.get('interactive', False)

    cmd = ['apt-get', 'install']
    env = os.environ.copy()
    if not interactive:
        env['DEBIAN_FRONTEND'] = 'noninteractive'
        cmd += ['-y']

    cmd += packages
    sudo(*cmd, env=env, direct=True)


def install_custom_packages(
        packages,
        **kwargs):

    def local_path(pkg):
        """
        apt-get is able to install local packages
        as long as they use full path or relative including ./
        """
        p = str(pkg)
        if p[0] not in '/\\.':
            return "./%s" % p
        return p

    interactive = kwargs.get('interactive', False)

    cmd = ['apt-get', 'install']
    env = os.environ.copy()
    if not interactive:
        env['DEBIAN_FRONTEND'] = 'noninteractive'
        cmd += ['-y']

    cmd += list(map(local_path, packages))
    sudo(*cmd, env=env, direct=True)


def install_build_deps(
        deps,
        **kwargs):
    """
    install debian build deps

    Debian Build-Depends can contain strings not handled by
    `apt-get install` such as "(>= 9~)"

    New `apt-get satisfy` command handles Build-Depends strings fine
    but it isn't available on current.

    Try to use `apt-get satisfy` if available,
    otherwise revert to stripping special strings and use `install`.
    """
    interactive = kwargs.get('interactive', False)

    if has_aptget_satisfy_():
        # unlike install, satisfy can handle versioned deps
        cmd = ['apt-get', 'satisfy']
        env = os.environ.copy()
        if not interactive:
            env['DEBIAN_FRONTEND'] = 'noninteractive'
            cmd += ['-y']
        cmd += deps
        sudo(*cmd, env=env, direct=True)
    else:
        # satisfy not available, strip special strings and use install
        packages = [strip_dep_(d) for d in deps]
        install_distro_packages(packages, **kwargs)


def get_build_deps_from_template(
        template_path,
        **kwargs):
    """
    parse Build-Depends from packaging template
    """
    distro = kwargs.get('distro')
    # render control file
    this_style = sys.modules[__name__]
    t = pkgtemplate.PackageTemplate(template_path, style=this_style)
    env = pkgtemplate.DUMMY_ENV.copy()
    if distro:
        env['distro'] = distro
    control_text = t.render_file_content('control', env=env)
    return get_build_deps_from_control_(control_text)


def get_build_deps_from_srcpkg(
        srcpkg_path,
        **_):
    """
    parse Build-Depends from source package
    """
    debar_path = get_srcpkg_debian_archive_(srcpkg_path.parent)
    log.info("unpacking debian archive: %s", debar_path)
    with tempfile.TemporaryDirectory(prefix='apkg_deb_') as td:
        unpack_path = unpack_archive(debar_path, td)
        control_path = unpack_path / 'control'
        control_text = control_path.open().read()
    return get_build_deps_from_control_(control_text)


# functions bellow with _ postfix are specific to this pkgstyle


def get_build_deps_from_control_(control_text):
    """
    parse Build-Depends from debian control file contents
    """
    m = re.findall(RE_BUILD_DEPENDS, control_text)
    if not m:
        msg = "unable to parse Build-Depends from control"
        raise ex.ParsingFailed(msg=msg)
    deps = []
    for deps_raw in m:
        deps += re.split(r'\s*,\s*', deps_raw)
    return deps


def get_srcpkg_debian_archive_(path):
    ars = glob.glob('%s/*.debian.tar.?z' % path)
    if not ars:
        msg = "unable to find debian archive in srcpkg: %s" % path
        raise ex.InvalidInput(msg=msg)
    if len(ars) > 1:
        msg = "multiple debian archives found in srcpkg: %s" % path
        raise ex.InvalidInput(msg=msg)
    return ars[0]


def has_aptget_satisfy_():
    """
    is `apt-get satisfy` command available?
    """
    o = run('apt-get', '-h', log_cmd=False, fatal=False)
    return 'satisfy' in o


def strip_dep_(dep):
    """
    strip special version strings (as found in Build-Depends)
    in order for dep to be installable through `apt-get install`
    """
    return re.split(r'[\s\[\(]', dep)[0]



import click

from apkg import adistro
from apkg.cli import cli
from apkg.pkgstyle import call_pkgstyle_fun
from apkg.commands.get_archive import get_archive
from apkg.commands.srcpkg import srcpkg as make_srcpkg
from apkg.util import common
from apkg.log import getLogger
from apkg.project import Project
from apkg.util.archive import unpack_archive


log = getLogger(__name__)


@cli.command(name="build-dep", aliases=['builddep'])
@click.argument('input_files', nargs=-1)
@click.option('-l', '--list', 'install', default=True, flag_value=False,
              help="list build deps only, don't install")
@click.option('-u', '--upstream', is_flag=True,
              help="use upstream template / archive / srcpkg")
@click.option('-s', '--srcpkg', is_flag=True,
              help="use source package")
@click.option('-a', '--archive', is_flag=True,
              help="use template (/build srcpkg) from archive")
@click.option('-d', '--distro',
              help="override target distro  [default: current]")
@click.option('-F', '--file-list', 'input_file_lists', multiple=True,
              help=("specify text file listing one input file per line"
                    ", use '-' to read from stdin"))
@click.option('--ask/--no-ask', 'interactive',
              default=False, show_default=True,
              help="enable/disable interactive mode")
# TODO: once py3.5 is dropped, add hidden=True
@click.option('-y', '--yes', 'interactive', flag_value=False,
              help="compat alias for --no-ask")
@click.help_option('-h', '--help',
                   help="show this help message")
def cli_build_dep(*args, **kwargs):

    """
    install or list build dependencies
    """
    deps = build_dep(*args, **kwargs)
    if not kwargs.get('install', True):
        common.print_results(deps)


def build_dep(
        upstream=False,
        srcpkg=False,
        archive=False,
        input_files=None,
        input_file_lists=None,
        install=True,
        distro=None,
        interactive=False,
        project=None):
    """
    parse and optionally install build dependencies

    pass install=False to only get list of deps without install

    returns list of build deps
    """
    action = 'installing' if install else 'listing'
    log.bold('%s build deps', action)

    proj = project or Project()
    distro = adistro.distro_arg(distro)
    log.info("target distro: %s", distro)

    # fetch pkgstyle (deb, rpm, arch, ...)
    template = proj.get_template_for_distro(distro)
    pkgstyle = template.pkgstyle

    infiles = common.parse_input_files(input_files, input_file_lists)

    if srcpkg:
        # use source package to determine deps
        if archive or not infiles:
            # build source package
            srcpkg_files = make_srcpkg(
                archive=archive,
                distro=distro,
                input_files=input_files,
                input_file_lists=input_file_lists,
                upstream=upstream,
                project=proj)
        else:
            # use specified source package
            srcpkg_files = infiles

        common.ensure_input_files(srcpkg_files)
        srcpkg_path = srcpkg_files[0]

        log.info("build deps from srcpkg: %s", srcpkg_path)
        deps = call_pkgstyle_fun(
            pkgstyle, 'get_build_deps_from_srcpkg',
            srcpkg_path)
    else:
        # use tempalte to determine deps
        if archive:
            archive_files = infiles

        if upstream:
            archive = True
            archive_files = get_archive(project=proj)

        if archive:
            common.ensure_input_files(archive_files)
            archive_path = archive_files[0]
            log.info("unpacking archive: %s", archive_path)
            unpack_path = unpack_archive(
                archive_path, proj.unpacked_archive_path)
            log.info("loading template from archive: %s", unpack_path)
            # load project with input_path from archive
            aproj = Project(path=unpack_path)
            template = aproj.get_template_for_distro(distro)

        log.info("build deps from template: %s", template.path)
        deps = call_pkgstyle_fun(
            pkgstyle, 'get_build_deps_from_template',
            template.path, distro=distro)

    if install:
        log.info("installing %s build deps...", len(deps))
        call_pkgstyle_fun(
            pkgstyle, 'install_build_deps',
            deps,
            distro=distro,
            interactive=interactive)

    return deps


APKG_CLI_COMMANDS = [cli_build_dep]



import sys

import click

from apkg import adistro
from apkg import ex
from apkg.log import getLogger
from apkg import pkgstyle


log = getLogger(__name__)


@click.command(name="install")
@click.argument('packages', nargs=-1)
@click.option('-D', '--distro-pkgs', is_flag=True,
              help="install packages from distro repos")
@click.option('-d', '--distro',
              help="override target distro  [default: current]")
@click.option('--ask/--no-ask', 'interactive',
              default=False, show_default=True,
              help="enable/disable interactive mode")
# TODO: once py3.5 is dropped, add hidden=True
@click.option('-y', '--yes', 'interactive', flag_value=False,
              help="compat alias for --no-ask")
@click.option('-F', '--file-list', 'input_file_lists', multiple=True,
              help=("specify text file listing one input package per line"
                    ", use '-' to read from stdin"))
@click.help_option('-h', '--help',
                   help="show this help message")
def cli_install(*args, **kwargs):
    """
    install packages using native package manager

    Supply one or more package to install and/or use -F <file-list>
    to read packages from specified file, one package per line.
    """
    return install(*args, **kwargs)


def install(
        packages=None,
        input_file_lists=None,
        distro_pkgs=False,
        distro=None,
        interactive=False):
    """
    install packages using native package manager
    """
    log.bold("installing packages")

    distro = adistro.distro_arg(distro)
    log.info("target distro: %s", distro)
    pkgs = packages
    for fn in input_file_lists:
        if fn == '-':
            f = sys.stdin
            # unable to get interactive input when piping stdin
            interactive = False
        else:
            f = open(fn)
        pkgs = pkgs + list(map(str.strip, f.readlines()))

    ps = pkgstyle.get_pkgstyle_for_distro(distro)
    if not ps:
        raise ex.DistroNotSupported(distro=distro)
    log.info("target pkgstyle: %s", ps.name)

    if distro_pkgs:
        result = ps.install_distro_packages(
            pkgs, distro=distro, interactive=interactive)
    else:
        result = ps.install_custom_packages(
            pkgs, distro=distro, interactive=interactive)

    log.success("installed %s packages", len(pkgs))

    return result


APKG_CLI_COMMANDS = [cli_install]



"""
apkg package style for **Arch** linux.

**source template:** `PKGBUILD`

**source package:** `PKGBUILD`

**packages:** `*.zst`
"""
import glob
from pathlib import Path
import sys

from apkg import ex
from apkg.log import getLogger
from apkg import pkgtemplate
from apkg.util.run import cd, run, sudo
import apkg.util.shutil35 as shutil


log = getLogger(__name__)


SUPPORTED_DISTROS = [
    'arch'
]


RE_PKG_NAME = r'pkgname\s*=\s*(\S+)'


def is_valid_template(path):
    pkgbuild = path / "PKGBUILD"
    return pkgbuild.exists()


def get_template_name(path):
    return parse_pkgbuild_(path / 'PKGBUILD', 'echo "$pkgname"')


def get_srcpkg_nvr(path):
    return parse_pkgbuild_(path, 'echo "$pkgname-$pkgver-$pkgrel"')


def build_srcpkg(
        build_path,
        out_path,
        archive_paths,
        template,
        env):
    archive_path = archive_paths[0]
    in_pkgbuild = build_path / 'PKGBUILD'
    out_pkgbuild = out_path / 'PKGBUILD'
    out_archive = out_path / archive_path.name
    log.info("building arch source package: %s", in_pkgbuild)
    template.render(build_path, env or {})
    out_path.mkdir(parents=True)
    log.info("copying PKGBUILD and archive to: %s", out_path)
    shutil.copyfile(in_pkgbuild, out_pkgbuild)
    shutil.copyfile(archive_path, out_archive)
    return [out_pkgbuild, out_archive]


def build_packages(
        build_path,
        out_path,
        srcpkg_paths,
        **_):
    srcpkg_path = srcpkg_paths[0]
    if srcpkg_path.name != 'PKGBUILD':
        raise ex.InvalidSourcePackageFormat(
            fmt="arch source package format is PKGBUILD but got: %s"
            % srcpkg_path.name)
    log.info("copying source package to build dir: %s", build_path)
    shutil.copytree(srcpkg_path.parent, build_path)
    # build package using makepkg
    log.info("starting arch package build using makepkg")
    with cd(build_path):
        run('makepkg', direct='auto')
    log.info("copying built packages to result dir: %s", out_path)
    out_path.mkdir(parents=True, exist_ok=True)
    pkgs = []
    # find and copy resulting packages
    for src_pkg in glob.iglob('%s/*.zst' % build_path):
        dst_pkg = out_path / Path(src_pkg).name
        shutil.copyfile(src_pkg, dst_pkg)
        pkgs.append(dst_pkg)

    return pkgs


def install_distro_packages(
        packages,
        **kwargs):
    interactive = kwargs.get('interactive', False)
    cmd = ['pacman', '-S']
    if not interactive:
        cmd += ['--noconfirm']
    cmd += packages
    sudo(*cmd, direct=True)


def install_custom_packages(
        packages,
        **kwargs):
    interactive = kwargs.get('interactive', False)
    cmd = ['pacman', '-U']
    if not interactive:
        cmd += ['--noconfirm']
    cmd += packages
    sudo(*cmd, direct=True)


def install_build_deps(
        deps,
        **kwargs):
    # no special handling for build deps on arch
    install_distro_packages(deps, **kwargs)


def get_build_deps_from_template(
        template_path,
        **kwargs):
    """
    parse depends from packaging template
    """
    distro = kwargs.get('distro')
    # render PKGBUILD
    this_style = sys.modules[__name__]
    t = pkgtemplate.PackageTemplate(template_path, style=this_style)
    env = pkgtemplate.DUMMY_ENV.copy()
    if distro:
        env['distro'] = distro
    pkgbuild_text = t.render_file_content('PKGBUILD', env=env)
    deps = parse_pkgbuild_content_(
        pkgbuild_text,
        'printf \'%s\n\' "${depends[@]}"')
    return deps.splitlines()


def get_build_deps_from_srcpkg(
        srcpkg_path,
        **_):
    """
    parse depends from source package (PKGBUILD)
    """
    deps = parse_pkgbuild_(
        srcpkg_path,
        'printf \'%s\n\' "${depends[@]}"')
    return deps.splitlines()


# functions bellow with _ postfix are specific to this pkgstyle


def parse_pkgbuild_(pkgbuild, bash):
    return run('bash', '-c', '. "%s" && %s'
               % (pkgbuild, bash), log_cmd=False)


def parse_pkgbuild_content_(pkgbuild_text, bash):
    pkgb = '%s\n%s' % (pkgbuild_text, bash)
    return run('bash', '-s', input=pkgb, log_cmd=False)



# -*- encoding: utf-8 -*-

import re
from packaging import version

from apkg.log import getLogger
from apkg.util.run import run


log = getLogger(__name__)


HTMLLISTPARSE_ERROR = None
try:
    import htmllistparse
except Exception as e:
    HTMLLISTPARSE_ERROR = e


RE_ARCHIVE_VERSION = r'[^\d]+-(.+)\.tar\..*'


def version_from_listing(html_listing_url):
    """
    get latest version from HTML listing
    """
    if HTMLLISTPARSE_ERROR:
        log.warning("unable to get upstream version from HTML listing: %s",
                    HTMLLISTPARSE_ERROR)
        return None
    log.verbose("getting upstream version from HTML listing: %s",
                html_listing_url)
    found = False
    v_max = version.parse('0')
    _, listing = htmllistparse.fetch_listing(
        html_listing_url, timeout=5)
    for f in listing:
        m = re.match(RE_ARCHIVE_VERSION, f.name)
        if not m:
            continue
        v = version.parse(m.group(1))
        v_max = max(v, v_max)
        found = True
    if found:
        return v_max
    return None


def version_from_script(script, script_name='script'):
    """
    get version from last stdout line of a script
    """
    log.verbose("getting upstream version from %s: %s", script_name, script)
    out = run(script)
    _, _, last_line = out.rpartition('\n')
    v = version.parse(last_line.strip())
    return v



"""
shared apkg testing functions
"""
# pylint: disable=dangerous-default-value
from pathlib import Path

import apkg.util.shutil35 as shutil


def init_testing_repo(repo_path, test_path, ignore_dirs=[]):
    dst = Path(test_path) / Path(repo_path).name
    inject_tree(repo_path, dst, ignore_dirs=ignore_dirs)
    return dst


def inject_tree(src_path, dst_path, ignore_dirs=[]):
    """
    copy all files from src_path into dst_path

    overwrite existing files
    """
    if not dst_path.exists():
        dst_path.mkdir(parents=True, exist_ok=True)

    # recursively copy all files
    for d, subdirs, files in shutil.walk(src_path):
        if ignore_dirs:
            # ignore selected dirs
            ignored = []
            for sd in subdirs:
                if sd in ignore_dirs:
                    ignored.append(sd)
            for sd_ignore in ignored:
                subdirs.remove(sd_ignore)

        rel_dir = Path(d).relative_to(src_path)
        dst_dir = dst_path / rel_dir
        dst_dir.mkdir(parents=True, exist_ok=True)

        for fn in files:
            dst = dst_dir / fn
            src = Path(d) / fn
            shutil.copy(src, dst)


def log_contains(string, caplog):
    for r in caplog.records:
        if string in r.message:
            return True
    return False



"""
apkg exceptions

when apkg CLI run results in exception being raised
its exit_code is returned as process exit code
"""


class ApkgException(Exception):
    msg_fmt = "An unknown error occurred"
    exit_code = 1

    def __init__(self, msg=None, exit_code=None, **kwargs):
        self.kwargs = kwargs
        if not msg:
            try:
                msg = self.msg_fmt % kwargs
            except Exception:
                # kwargs doesn't match those in message.
                # Returning this is still better than nothing.
                msg = self.msg_fmt
        if exit_code is not None:
            self.exit_code = exit_code
        super().__init__(msg)


# 10-29: Invalid
class InvalidUsage(ApkgException):
    msg_fmt = "Invalid usage: %(fail)s"
    exit_code = 10


class InvalidInput(ApkgException):
    msg_fmt = "Invalid input: %(fail)s"
    exit_code = 11


class InvalidType(ApkgException):
    msg_fmt = "Invalid type: $(var) must be %(desired)s but is %(type)s"
    exit_code = 12


class InvalidChoice(ApkgException):
    msg_fmt = "Invalid choice: %(var)s must be one of: %(opts)s (is: %(val)s)"
    exit_code = 13


class InvalidVersion(ApkgException):
    msg_fmt = "Invalid version: %(ver)s"
    exit_code = 14


class InvalidFormat(ApkgException):
    msg_fmt = "Invalid format: %(fmt)s"
    exit_code = 16


class InvalidArchiveFormat(InvalidFormat):
    msg_fmt = "Invalid archive format: %(fmt)s"
    exit_code = 17


class InvalidSourcePackageFormat(InvalidFormat):
    msg_fmt = "Invalid source package format: %(fmt)s"
    exit_code = 18


# 30-39: Missing and NotFound
class MissingRequiredArgument(ApkgException):
    msg_fmt = "Missing required argument: %(arg)s"
    exit_code = 30


class MissingRequiredConfigOption(ApkgException):
    msg_fmt = "Missing required config option: %(opt)s"
    exit_code = 31


class MissingPackagingTemplate(ApkgException):
    msg_fmt = "Missing package template: %(temp)s"
    exit_code = 32


class ArchiveNotFound(ApkgException):
    msg_fmt = "%(type)s archive not found: %(ar)s"
    exit_code = 36


class SourcePackageNotFound(ApkgException):
    msg_fmt = "%(type)s source package not found: %(srcpkg)s"
    exit_code = 37


# 40-49: Parsing
class ParsingFailed(ApkgException):
    msg_fmt = "Unable to parse: %(fail)s"
    exit_code = 42


# 50-59: remote failures
class FileDownloadFailed(ApkgException):
    msg_fmt = "Failed to download file with code %(code)s:\n\n%(url)s"
    exit_code = 52


# 60-69: Command
class CommandNotFound(ApkgException):
    msg_fmt = "Command not found: %(cmd)s"
    exit_code = 60


class CommandFailed(ApkgException):
    msg_fmt = "Command failed: %(cmd)s"
    exit_code = 62


class UnexpectedCommandOutput(ApkgException):
    msg_fmt = "Unexpected command output: %(out)s"
    exit_code = 64


# 70-79: Distro
class DistroNotSupported(ApkgException):
    msg_fmt = "Distro not supported: %(distro)s"
    exit_code = 70


# 80-89: auto-magic
class UnableToDetectUpstreamVersion(ApkgException):
    msg_fmt = (
        "Unable to detect upstream version.\n\n"
        "Please consider one of following:\n\n"
        "1) set upstream.archive_url "
        "(requires htmllistparse module)\n"
        "2) set upstream.version_script to custom script\n"
        "3) manually supply version using -v/--version option")
    exit_code = 84



"""
apkg archive (tarball) utils
"""
from pathlib import Path

from apkg.parse import split_archive_fn, parse_version
import apkg.util.shutil35 as shutil

from apkg import ex


def unpack_archive(archive_path, out_path):
    """
    unpack supplied archive into out_path dir

    archive is expected to contain a single root dir

    return path to extracted root dir
    """
    out_path = Path(out_path)
    out_path.mkdir(parents=True, exist_ok=True)
    root_files_old = set(out_path.glob("*"))
    # shutil doesn't provide a way to check extracted files :(
    # pautil has bugs and got last release in 2016...
    shutil.unpack_archive(archive_path, out_path)
    root_files_new = set(out_path.glob("*"))
    root_files = root_files_new - root_files_old
    n_root_files = len(root_files)
    if n_root_files != 1:
        fmt = "Expected a single root dir but instead got %d files in root"
        raise ex.InvalidArchiveFormat(fmt=fmt % n_root_files)
    return root_files.pop()


def get_archive_version(archive_path):
    """
    return archive version detected from archive name
    """
    archive_path = Path(archive_path)
    _, _, ver, _ = split_archive_fn(archive_path.name)
    return parse_version(ver)



"""
apkg package style for RPM-based distros
such as Fedora, CentOS, SUSE, RHEL.

**source template**: `*.spec` and friends

**source package:** `*.src.rpm`

**packages:** `*.rpm`

**required distro packages**:

 * core: `rpm-build`
 * isolated build: `mock`
"""
import glob
from pathlib import Path
import re
import subprocess
import sys

from apkg import ex
from apkg.util import common
from apkg.log import getLogger
from apkg import pkgtemplate
from apkg.util.run import run, sudo
import apkg.util.shutil35 as shutil


log = getLogger(__name__)


SUPPORTED_DISTROS = [
    "fedora",
    "centos",
    "rhel",
    "oracle",
    "pidora",
    "opensuse",
    "scientific",
]
DISTRO_REQUIRES = {
    'core': ['rpm-build'],
    'isolated': ['mock'],
}


RE_PKG_NAME = r'Name:\s*(\S+)'
RE_BUILD_REQUIRES = r'BuildRequires:\s*(.*)'
RE_RPMBUILD_OUT_RPM = r'Wrote:\s+(.*\.rpm)\s*'
RE_RPMBUILD_OUT_SRPM = r'Wrote:\s+(.*\.src.rpm)\s*'


def is_valid_template(path):
    return bool(get_spec_(path))


def get_template_name(path):
    spec = get_spec_(path)

    for line in spec.open():
        m = re.match(RE_PKG_NAME, line)
        if m:
            return m.group(1)

    raise ex.ParsingFailed(
        msg="unable to determine Name from: %s" % spec)


def get_srcpkg_nvr(path):
    name = str(path.name)
    m = re.match(r'(.*)\.src.rpm', name)
    if m:
        return m.group(1)
    return name


def build_srcpkg(
        build_path,
        out_path,
        archive_paths,
        template,
        env):
    """
    build .src.rpm source package
    """
    rpmbuild_topdir = build_path / 'rpmbuild'
    rpmbuild_src = rpmbuild_topdir / 'SOURCES'
    rpmbuild_spec = rpmbuild_topdir / 'SPEC'

    rpmbuild_src.mkdir(parents=True, exist_ok=True)

    template.render(rpmbuild_src, env)

    spec_src_path = get_spec_(rpmbuild_src)
    spec_path = rpmbuild_spec / spec_src_path.name
    log.verbose("moving .spec file into SPEC: %s", spec_path)
    rpmbuild_spec.mkdir(exist_ok=True)
    spec_src_path.rename(spec_path)

    log.info("copying archive files into SOURCES: %s", rpmbuild_src)
    for src_path in archive_paths:
        dst_path = rpmbuild_src / src_path.name
        shutil.copyfile(src_path, dst_path)
    log.info("building .src.rpm using rpmbuild")
    out = run('rpmbuild', '-bs',
              '--define', '_topdir %s' % rpmbuild_topdir.resolve(),
              spec_path)

    log.info("copying .src.rpm to result dir: %s", out_path)
    out_path.mkdir(parents=True)
    srcpkgs = []
    for m in re.finditer(RE_RPMBUILD_OUT_SRPM, out):
        srpm = m.group(1)
        src_srpm = Path(srpm)
        dst_srpm = out_path / src_srpm.name
        shutil.copyfile(src_srpm, dst_srpm)
        srcpkgs.append(dst_srpm)
    if not srcpkgs:
        raise ex.ParsingFailed(
            msg="unable to parse rpmbuild results")
    return srcpkgs


def build_packages(
        build_path,
        out_path,
        srcpkg_paths,
        **kwargs):
    """
    build .rpm packages from .src.rpm
    """
    isolated = kwargs.get('isolated')
    pkgs = []
    srcpkg_path = srcpkg_paths[0]

    if isolated:
        log.info("starting isolated .rpm build using mock")
        # sudo shouldn't be necessary when in mock group
        # but apkg can't rely on that especially in containers etc.
        sudo('mock',
             '--resultdir', build_path,
             srcpkg_path,
             preserve_env=True,
             direct='auto')
        log.info("copying built packages to result dir: %s", out_path)
        out_path.mkdir(parents=True)
        for rpm in glob.iglob('%s/*.rpm' % build_path):
            src_pkg = Path(rpm)
            dst_pkg = out_path / src_pkg.name
            shutil.copyfile(src_pkg, dst_pkg)
            pkgs.append(dst_pkg)
    else:
        log.info("starting direct host .rpm build using rpmbuild")
        rpmbuild_topdir = build_path / 'rpmbuild'
        rpmbuild_topdir.mkdir(parents=True, exist_ok=True)
        out = run('rpmbuild', '--rebuild',
                  '--define', '_topdir %s' % rpmbuild_topdir.resolve(),
                  srcpkg_path)
        log.info("copying built packages to result dir: %s", out_path)
        out_path.mkdir(parents=True)
        for m in re.finditer(RE_RPMBUILD_OUT_RPM, out):
            rpm = m.group(1)
            src_pkg = Path(rpm)
            dst_pkg = out_path / src_pkg.name
            shutil.copyfile(src_pkg, dst_pkg)
            pkgs.append(dst_pkg)
        if not pkgs:
            raise ex.ParsingFailed(
                msg="unable to parse rpmbuild results")

    return pkgs


def install_build_deps(
        deps,
        **kwargs):
    # dnf/zypper install handles build deps
    install_distro_packages(deps, **kwargs)


def install_custom_packages(
        packages,
        **kwargs):
    # dnf/zypper install handles local packages
    install_distro_packages(packages, **kwargs)


def install_distro_packages(
        packages,
        **kwargs):
    interactive = kwargs.get('interactive', False)
    distro = kwargs.get('distro')
    pm = get_package_manager_(distro)

    cmd = [pm, 'install']
    if pm == 'zypper':
        # use zypper capabilities
        cmd += ['-C']
    if not interactive:
        cmd += ['-y']
    cmd += packages
    sudo(*cmd, direct=True)


def get_build_deps_from_template(
        template_path,
        **kwargs):
    """
    parse BuildRequires from packaging template
    """
    distro = kwargs.get('distro')
    spec_path = get_spec_(template_path).relative_to(template_path)
    # render .spec file
    this_style = sys.modules[__name__]
    t = pkgtemplate.PackageTemplate(template_path, style=this_style)
    env = pkgtemplate.DUMMY_ENV.copy()
    if distro:
        env['distro'] = distro
    spec_text = t.render_file_content(spec_path, env=env)
    return get_build_deps_from_spec_(spec_text)


def get_build_deps_from_srcpkg(
        srcpkg_path,
        **_):
    """
    parse BuildRequires from .src.rpm
    """
    cmd = "rpm2cpio '%s' | cpio -i --to-stdout '*.spec'" % srcpkg_path
    spec_text = subprocess.getoutput(cmd)
    return get_build_deps_from_spec_(spec_text)


# functions bellow with _ postfix are specific to this pkgstyle


def get_package_manager_(distro):
    default = 'dnf'
    if not distro:
        return default
    if distro.startswith('opensuse'):
        return 'zypper'
    m = re.match(r'(?:centos|rhel|oracle|scientific)-(\d+)', distro)
    if m and int(m.group(1)) <= 7:
        # use yum on EL <= 7
        return 'yum'
    return default


def get_spec_(path):
    for s in glob.iglob("%s/*.spec" % path):
        return Path(s)
    return None


def get_build_deps_from_spec_(spec_text):
    """
    parse BuildRequires from .spec file content
    """
    # parse spec to expand macros
    # done through temp file because rpmspec doesn't work
    # on /dev/stdin on openSUSE for some reason :-/
    with common.text_tempfile(spec_text, prefix='apkg_rpm.spec_') as spec_path:
        spec_parsed = run('rpmspec', '-P', spec_path)

    return re.findall(RE_BUILD_REQUIRES, spec_parsed)



# -*- encoding: utf-8 -*-
from contextlib import contextmanager
import os
import subprocess
import sys

from apkg import ex
from apkg.log import getLogger, T, LOG_LEVEL, INFO


log = getLogger(__name__)


IS_ROOT = False
if os.geteuid() == 0:
    IS_ROOT = True


def log_cmd_fail(cmd, cout):
    log.error("command failed: %s", T.command(cmd))
    if cout:
        log.bold("stdout:")
        log.info(cout)
    if cout.stderr:
        log.bold("stderr:")
        log.info(cout.stderr)


def run(*cmd, **kwargs):
    """
    run system commands easily - a subprocess.Popen wrapper

    run('echo', 'hello world')
    """
    fatal = kwargs.get('fatal', True)
    direct = kwargs.get('direct', False)
    silent = kwargs.get('silent', False)
    log_cmd = kwargs.get('log_cmd', True)
    log_fail = kwargs.get('log_fail', True)
    log_fun = kwargs.get('log_fun', log.command)
    _input = kwargs.get('input')
    print_stdout = kwargs.get('print_stdout', False)
    print_stderr = kwargs.get('print_stderr', False)
    print_output = kwargs.get('print_output', False)
    env = kwargs.get('env', None)

    # TODO: escape parameters with whitespace
    cmd = [str(c) for c in cmd]
    cmd_str = ' '.join(cmd)

    if silent:
        log_cmd = False
        log_fail = False
        print_stdout = False
        print_stderr = False

    if print_output:
        print_stdout = True
        print_stderr = True

    if _input:
        stdin = subprocess.PIPE
        if not isinstance(_input, bytes):
            _input = bytes(_input, 'utf-8')
    else:
        stdin = None

    if direct == 'auto':
        direct = bool(sys.stdout.isatty() and LOG_LEVEL <= INFO)
    if direct:
        stdout = None
        stderr = None
    else:
        stdout = subprocess.PIPE
        stderr = subprocess.PIPE

    if log_cmd:
        log_fun(cmd_str)

    try:
        prc = subprocess.Popen(cmd, stdin=stdin, stdout=stdout,
                               stderr=stderr, env=env)
    except OSError:
        raise ex.CommandNotFound(cmd=cmd[0])
    out, err = prc.communicate(input=_input)

    if isinstance(out, bytes):
        out = out.decode('utf-8')
    if isinstance(err, bytes):
        err = err.decode('utf-8')

    if out:
        out = out.rstrip()
        if print_stdout:
            log.info(out)
    else:
        out = ''

    if err:
        err = err.rstrip()
        if print_stderr:
            log.info(err)
    else:
        err = ''

    cout = CommandOutput(out)
    cout.stderr = err
    cout.return_code = prc.returncode
    cout.cmd = cmd_str
    if prc.returncode != 0:
        if log_fail:
            log_cmd_fail(cmd_str, cout)
        if fatal:
            raise ex.CommandFailed(cmd=cmd, out=cout)
    return cout


def sudo(*cmd, **kwargs):
    preserve_env = kwargs.pop('preserve_env', False)
    if 'env' in kwargs:
        preserve_env = True
    if not IS_ROOT:
        sudo_cmd = ['sudo']
        if preserve_env:
            sudo_cmd.append('-E')
        cmd = sudo_cmd + list(cmd)
        kwargs['log_fun'] = log.sudo
    return run(*cmd, **kwargs)


@contextmanager
def cd(newdir):
    """
    Temporarily change current directory.
    """
    olddir = os.getcwd()
    oldpwd = os.environ['PWD']
    newpath = os.path.abspath(os.path.expanduser(str(newdir)))
    os.chdir(newpath)
    os.environ['PWD'] = newpath
    try:
        yield
    finally:
        os.chdir(olddir)
        os.environ['PWD'] = oldpwd


class ShellCommand:
    command = None

    def __init__(self):
        if self.command is None:
            self.command = self.__class__.__name__.lower()

    def __call__(self, *params, **kwargs):
        return run(self.command, *params, **kwargs)


class CommandOutput(str):
    """
    Just a string subclass with attribute access.
    """
    cmd = None
    stderr = None
    return_code = None

    @property
    def success(self):
        return self.return_code == 0



#!/usr/bin/env python3
import setuptools


def get_version_and_cmdclass(package_path):
    """Load version.py module without importing the whole package.

    Template code from miniver
    """
    import os
    from importlib.util import module_from_spec, spec_from_file_location

    spec = spec_from_file_location(
        "version", os.path.join(package_path, "_version.py"))
    module = module_from_spec(spec)
    spec.loader.exec_module(module)
    return module.__version__, module.cmdclass


version, cmdclass = get_version_and_cmdclass("apkg")


setuptools.setup(
    name='apkg',
    version=version,
    cmdclass=cmdclass,
)



import os

import click

from apkg import adistro
from apkg.log import getLogger, T
from apkg.project import Project


log = getLogger(__name__)


@click.command(name='status')
@click.help_option('-h', '--help', help='show this help')
def cli_status():
    """
    show status of current project
    """
    status()


def status():
    """
    show status of current project
    """
    proj = Project()

    msg = "project name:            {t.bold}{name}{t.normal}"
    print(msg.format(name=proj.name, t=T))
    msg = "project base path:       {t.bold}{path}{t.normal}"
    print(msg.format(path=proj.path.resolve(), t=T))

    msg = "project VCS:             {t.bold}{vcs}{t.normal}"
    print(msg.format(vcs=proj.vcs or 'none', t=T))

    msg = "project config:          {t.bold}{path}{t.normal}"
    if proj.config_path.exists():
        msg += " ({t.green}exists{t.normal})"
    else:
        msg += " ({t.warn}doesn't exist{t.normal})"
    print(msg.format(path=proj.config_path, t=T))

    msg = "package templates path:  {t.bold}{path}{t.normal}"
    if proj.templates_path.exists():
        msg += " ({t.green}exists{t.normal})"
    else:
        msg += " ({t.red}doesn't exist{t.normal})"
    print(msg.format(path=proj.templates_path, t=T))

    print("package templates:")
    if proj.templates:
        msg_lines = []
        for template in proj.templates:
            short_path = os.path.join(*list(template.path.parts)[-3:])
            msg_lines.append("    {t.green}%s{t.normal}: {t.bold}%s{t.normal}"
                             % (template.pkgstyle.name, short_path))
        msg = "\n".join(msg_lines)
    else:
        msg = "    {t.red}no package templates found{t.normal}"
    print(msg.format(dir=proj.templates_path, t=T))

    print()
    # distro status
    msg = "current distro: {t.cyan}{id}{t.normal} / {t.cyan}{full}{t.normal}"
    print(msg.format(full=adistro.fullname(), id=adistro.idver(), t=T))

    # we know what we're doing here
    # pylint: disable=protected-access
    template = proj._get_template_for_distro(adistro.idver())
    msg = "    package style: "
    if template:
        style = template.pkgstyle.name
        msg += "{t.green}%s{t.normal}" % style
    else:
        msg += "{t.warn}unsupported{t.normal}"
    print(msg.format(t=T))
    msg = "    package template: "
    if template:
        msg += "{t.green}%s{t.normal}" % template.path
    else:
        msg += "{t.warn}unsupported{t.normal}"
    print(msg.format(t=T))


APKG_CLI_COMMANDS = [cli_status]



"""
A wrapper around selected shutil and os functions in order
to ensure compatibility with pathlib.Path in Python 3.5.

TODO: Once Python 3.5 support is dropped, replace all

    import apkg.util.shutil35 as shutil

with

    import shutil

and remove this compat module.
"""
import shutil
import os


def copy(src, dst, **kwargs):
    return shutil.copy(str(src), str(dst), **kwargs)


def copyfile(src, dst, **kwargs):
    return shutil.copyfile(str(src), str(dst), **kwargs)


def copytree(src, dst, **kwargs):
    return shutil.copytree(str(src), str(dst), **kwargs)


def rmtree(path, **kwargs):
    return shutil.rmtree(str(path), **kwargs)


def walk(top, **kwargs):
    return os.walk(str(top), **kwargs)


# this function copies upstream arg names 1:1
# pylint: disable=redefined-builtin
def unpack_archive(filename, extract_dir=None, format=None):
    # shutil supports pathlib starting with python 3.7
    if extract_dir:
        extract_dir = str(extract_dir)
    return shutil.unpack_archive(str(filename), extract_dir, format)



"""
module for handling and rendering apkg package templates
"""
from pathlib import Path
import re

import jinja2

from apkg.log import getLogger
from apkg import pkgstyle as _pkgstyle
import apkg.util.shutil35 as shutil


log = getLogger(__name__)


DUMMY_ENV = {
    'name': 'PKGNAME',
    'version': '0.VERSION',
    'release': '0.RELEASE',
    'nvr': 'NVR',
    'distro': 'DISTRO',
}


def default_render_filter(path):
    if str(path).endswith('.patch'):
        return False
    return True


class PackageTemplate:
    def __init__(self, path, style=None):
        self.path = Path(path)
        self.style = style

    @property
    def pkgstyle(self):
        if not self.style:
            self.style = _pkgstyle.get_pkgstyle_for_template(self.path)
        return self.style

    def render(self, out_path, env,
               render_filter=default_render_filter,
               includes=None, excludes=None):
        """
        render package template into specified output directory

        Args:
            out_path: output base path
            env: vars available from template
            render_filter: function to determine which files need rendering
            includes: render only files matching these regexes
            excludes: don't render any files matching these regexes
        """
        def is_included(fn):
            if includes:
                for inc_re in includes:
                    if re.match(inc_re, fn):
                        break
                else:
                    return False
            if excludes:
                for exc_re in excludes:
                    if re.match(exc_re, fn):
                        return False
            return True

        log.info("renderding package template: %s -> %s", self.path, out_path)
        if out_path.exists():
            log.verbose("template render dir exists: %s", out_path)
        else:
            out_path.mkdir(parents=True, exist_ok=True)

        # recursively render all files
        for d, _, files in shutil.walk(self.path):
            rel_dir = Path(d).relative_to(self.path)
            dst_dir = out_path / rel_dir
            dst_dir.mkdir(parents=True, exist_ok=True)

            for fn in files:
                dst = out_path / rel_dir / fn
                src = Path(d) / fn
                if not is_included(fn):
                    log.verbose("file excluded from render: %s", fn)
                    continue

                # TODO: filtering should be exposed through config
                if render_filter(src):
                    log.verbose("rendering file: %s -> %s", src, dst)
                    t = None
                    with src.open('r') as srcf:
                        t = jinja2.Template(srcf.read())
                    with dst.open('w') as dstf:
                        dstf.write(t.render(**env) + '\n')
                else:
                    log.verbose(
                        "copying file without render: %s -> %s", src, dst)
                    shutil.copyfile(src, dst)
                # preserve original permission
                dst.chmod(src.stat().st_mode)

    def render_file_content(self, name, env):
        """
        render template file in memory and return its content
        """
        src = self.path / name
        with src.open('r') as srcf:
            t = jinja2.Template(srcf.read())
        return t.render(**env) + '\n'



from contextlib import contextmanager
from pathlib import Path
import sys
import tempfile

from apkg import ex
from apkg.log import getLogger
import apkg.util.shutil35 as shutil


log = getLogger(__name__)


def copy_paths(paths, dst):
    """
    utility to copy a list of paths to dst
    """
    if not dst.exists():
        dst.mkdir(parents=True, exist_ok=True)
    dst_full = dst.resolve()
    new_paths = []
    for p in paths:
        if p.parent.resolve() == dst_full:
            new_paths.append(p)
        else:
            p_dst = dst / p.name
            log.verbose("copying file: %s -> %s", p, p_dst)
            shutil.copy(p, p_dst)
            new_paths.append(p_dst)
    return new_paths


def get_cached_paths(proj, cache_name, cache_key,
                     result_dir=None):
    """
    get cached files and move them to result_dir if specified
    """
    paths = proj.cache.get(cache_name, cache_key)
    if not paths:
        return None
    if result_dir:
        paths = copy_paths(paths, Path(result_dir))
    return paths


def print_results(results):
    """
    print results received from apkg command
    """
    try:
        for r in results:
            print(str(r))
    except TypeError:
        print(str(results))


def parse_input_files(files, file_lists):
    """
    utility to parse apkg input files and input file lists
    into a single list of input files
    """
    if not files:
        files = []
    if not file_lists:
        file_lists = []

    all_files = [Path(f) for f in files]

    if len([fl for fl in file_lists if fl == '-']) > 1:
        fail = "requested to read stdin multiple times"
        raise ex.InvalidInput(fail=fail)

    for fl in file_lists:
        if fl == '-':
            f = sys.stdin
        else:
            f = open(fl, 'r')
        all_files += [Path(ln.strip()) for ln in f.readlines()]
        f.close()

    return all_files


def ensure_input_files(infiles):
    if not infiles:
        raise ex.InvalidInput(
            fail="no input file(s) specified")
    for f in infiles:
        if not f or not f.exists():
            raise ex.InvalidInput(
                fail="input file not found: %s" % f)


@contextmanager
def text_tempfile(text, prefix='apkg_tmp_'):
    """
    write text to a new temporary file and return its path

    file is deleted after use
    """
    f = tempfile.NamedTemporaryFile(
        prefix=prefix, mode='w+t', delete=False)
    path = Path(f.name)
    f.write(text)
    f.close()
    try:
        yield path
    finally:
        path.unlink()
